from re import match, compile
from enum import Enum
from collections import namedtuple
from pythautomata.base_types.alphabet import Alphabet
from random import randrange, random, choice, seed
from pythautomata.regular_expressions.regular_expression import RegularExpression

class ProductionRules(Enum):
    CONCATENATION = ""
    CHOICE = "|"
    KLEENE_STAR = "*"


RuleApplication = namedtuple(
    'RuleApplication', 'rule dependencies')


class RegularExpressionGenerator():

    __possible_applications = (RuleApplication(ProductionRules.CONCATENATION, (None, None)),
                               RuleApplication(ProductionRules.CHOICE, (None, None)),
                               RuleApplication(ProductionRules.KLEENE_STAR, (None,)))

    def __init__(self, alphabet: Alphabet):        
        self._alphabet = alphabet
        seed()
    
    
    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    def generate_regular_expression_with(self,iterations: int, some_seed:int = 42):
        seed(some_seed)
        if self.alphabet is None or len(self.alphabet) < 2:
            raise ValueError("Alphabet must not be None, and have at least two symbols")

        expression = None
        for _ in range(iterations):
            expression = self.__add_production_rule_to(expression)
        result = self.__convert_to_string(expression, list(self.alphabet.symbols))
        pattern = None
        if iterations > 0 and expression.rule == ProductionRules.CHOICE:
             pattern = compile(f"^(?:{result[0]})$")
        pattern = compile(f"^{result[0]}$")
        return RegularExpression(alphabet = self.alphabet, pattern = pattern)

    def __add_production_rule_to(self, result, noKleeneStar=False):
        if result is None:
            ruleToAdd = self.__choose_random_rule_with(noKleeneStar)
            return RuleApplication(ruleToAdd.rule, list(ruleToAdd.dependencies))
        else:
            index = randrange(len(result.dependencies))
            selected = result.dependencies[index]
            noKleeneStar = result.rule == ProductionRules.KLEENE_STAR or \
                noKleeneStar and result.rule == ProductionRules.CHOICE
            result.dependencies[index] = self.__add_production_rule_to(selected, noKleeneStar)
            return result

    def __choose_random_rule_with(self, noKleeneStar):
        roll = random()
        if noKleeneStar:
            index = 0 if roll < 0.66666 else 1
        else:
            index = 0 if roll < 0.556 else 1 if roll < 0.834 else 2
        return self.__possible_applications[index]

    def __convert_to_string(self, result, symbols: list):
        if result is None:
            return str(choice(symbols)), None

        symbol = result.rule.value
        if result.rule == ProductionRules.KLEENE_STAR:
            child, childRule = self.__convert_to_string(result.dependencies[0], symbols)
            substring = self.__get_substring_for(child, result.rule, childRule)
            return f"{substring}{symbol}?", result.rule
        else:
            left, leftRule = self.__convert_to_string(result.dependencies[0], symbols)
            right, rightRule = self.__convert_to_string(result.dependencies[1], symbols)
            leftSubstring = self.__get_substring_for(left, result.rule, leftRule)
            rightSubstring = self.__get_substring_for(right, result.rule, rightRule)
            return f"{leftSubstring}{symbol}{rightSubstring}", result.rule

    def __get_substring_for(self, child: str, rule: ProductionRules, child_rule: ProductionRules):
        if rule == ProductionRules.CONCATENATION and child_rule == ProductionRules.CHOICE or \
                rule == ProductionRules.KLEENE_STAR and child_rule is not None:
            return f"(?:{child})"
        return child


from pythautomata.base_types.guard import Guard
from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.symbol import SymbolInfinity as Inf
from pythautomata.base_types.symbol import SymbolNegativeInfinity as NInf
from pythautomata.base_types.symbolic_state import SymbolicState
from pythautomata.boolean_algebra_learner.boolean_algebra_learner import \
    BooleanAlgebraLearner
from pythautomata.guards.closed_interval_guard import ClosedIntervalGuard
from pythautomata.guards.union_guard import UnionGuard

inf = Inf()
ninf = NInf()
class ClosedDiscreteIntervalLearner(BooleanAlgebraLearner):

    
    def learn(self, multidict: dict[SymbolicState, list[Symbol]]) -> list[tuple[Guard, SymbolicState]]:
        if(len(multidict) == 0):
            return []
        ret:list[tuple[Guard, SymbolicState]] = []
        dict_state_to_symbol_intervals = self._init_dict_state_symbol_intervals(multidict)
        state_intervals:dict[SymbolicState,list[list[Symbol]]] = {}
        #TODO repensarlo para intervalos abierto-cerrado o vice versa
        intervals:list[list[Symbol]]
        for state, intrs in dict_state_to_symbol_intervals.items():
            intervals = [intr.copy() for intr in intrs]
            others:list[Symbol] = []
            for st, syms_l in dict_state_to_symbol_intervals.items():
                if st.name != state.name:
                    for l in syms_l:
                        others.extend(l)
            others.sort()
            
            new_intervals:list[list[Symbol]] = []
            init_len = len(intervals)
            is_fixed_point = False
            while len(intervals):
                if len(others) == 0:
                    break
                if is_fixed_point:
                    break
                is_fixed_point = True
                right = intervals[0]
                #right because I will split the interval in left of other and right of other
                for other in others:
                    if right[0] > other:
                        # then @other is smaller than every symbol in right
                        continue 
                    if right[-1] < other:
                        # then @other and the rest of the symbols in @others are bigger than those in the right interval
                        break
                    is_fixed_point = False
                    left = []
                    while True:
                        s=right[0]
                        #if @s is bigger than @other then i've found the symbol in which i should break the interval
                        if s > other:
                            #the interval will be between the first element of left and the symbol that comes before @other
                            other_copy = other.copy()
                            other_copy.add_to_value(-1)
                            left.append(other_copy)
                            new_intervals.append(left)
                            break
                        #if not, should remove s form right and add it to right
                        left.append(right.pop(0))
            new_intervals.extend(intervals)
            assert(len(new_intervals)>=init_len)
            state_intervals[state]=new_intervals

            
        
        self._add_infinities_to_state_intervals(state_intervals)
        for state, intervals in state_intervals.items():
            interval_guards = []
            for interval in intervals:
                assert(len(interval))
                interval_guards.append(ClosedIntervalGuard(interval[0],interval[-1]))
            guard = UnionGuard(*interval_guards)
            ret.append((guard, state))
        
        return ret

    def _add_infinities_to_state_intervals(self, state_intervals: dict[SymbolicState, list[list[Symbol]]]):
        min_state:SymbolicState
        min:Symbol = inf
        max_state:SymbolicState
        max:Symbol = ninf
        for state, intervals in state_intervals.items():
            assert(len(intervals))
            if min > intervals[0][0]:
                min = intervals[0][0]
                min_state = state
            if max < intervals[-1][-1]:
                max = intervals[-1][-1]
                max_state = state
        state_intervals[min_state][0].insert(0, ninf)
        state_intervals[max_state][-1].append(inf)


    def _init_dict_state_symbol_intervals(self, multidict: dict[SymbolicState, list[Symbol]]):
        dict_state_to_symbol_intervals:dict[SymbolicState, list[list[Symbol]]] = {}

        for state, symbols in multidict.items():
            symbols.sort()
            dict_state_to_symbol_intervals[state] = [symbols]
        return dict_state_to_symbol_intervals



from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.moore_state import MooreState
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.exceptions.model_importing_error import ModelImportingError
from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton
from pythautomata.model_comparators.moore_machine_comparison_strategy import \
    MooreMachineComparisonStrategy as MMComparator

import pydot


class StandardDotMMNotationImporter:

    @staticmethod
    def import_automata(path):
        graph = pydot.graph_from_dot_file(path)[0]
        name = graph.obj_dict['name']
        nodes = {}
        symbols = set()
        output_symbols = set()
        initial_state = None
        for node in graph.get_node_list():
            if node.get_name() not in ['__start0', '', '"\\n"']:
                value = SymbolStr(node.get_attributes()['label'].split("| ")[1][:-3])
                if value not in output_symbols:
                    output_symbols.add(value)
                nodes[node.get_name()] = MooreState(node.get_name(), value)

        for edge in graph.get_edge_list():
            if edge.get_source() == '__start0':
                initial_state = nodes[edge.get_destination()]
            else:
                edge_label = edge.get_attributes()['label']
                if edge_label[0] == "\"" and edge_label[0] == edge_label[-1]:
                    edge_label = edge_label[1:-1]
                symbol = SymbolStr(edge_label)
                symbols.add(symbol)
                nodes[edge.get_source()].add_transition(
                    symbol, nodes[edge.get_destination()])

        states = nodes.values()
        alphabet = Alphabet(frozenset(symbols))
        output_alphabet = Alphabet(frozenset(output_symbols))
        if initial_state is None:
            raise ModelImportingError

        return MooreMachineAutomaton(alphabet, output_alphabet, initial_state, states, MMComparator(), name)



from pythautomata.base_types.state import State
from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA


class DFAMinimizer():

    def __init__(self, dfa: DFA):
        self.dfa = dfa

    def minimize(self) -> DFA:
        """
        Minimizes DFA.

        Returns
        -------
        DFA
            A minimized DFA
        """
        final_eq_class, equivalence_class_0 = self._get_final_eq_class()

        if self._is_minimizable(final_eq_class):
            minimalDfa = self._convert_to_dfa(
                final_eq_class, equivalence_class_0)
            minimalDfa._exporting_strategies = self.dfa._exporting_strategies
            return minimalDfa
        else:
            return self.dfa

    def _get_final_eq_class(self):
        final_states = self._get_final_states()
        non_final_states = self._get_non_final_states()
        equivalence_class_0 = [final_states, non_final_states]
        equivalence_classes = [equivalence_class_0]
        current_iteration = 0
        finished = False

        while not finished:
            new_eq_class = self._next_eq_class_from(
                equivalence_classes[current_iteration])
            current_iteration += 1
            equivalence_classes.append(new_eq_class)
            finished = self._eq_class_iterations_finished(
                equivalence_classes, current_iteration)
        final_eq_class = equivalence_classes[current_iteration]
        return final_eq_class, equivalence_class_0

    def _is_minimizable(self, final_eq_class) -> bool:
        minimizable = not len(final_eq_class) == len(
            [state for state in self.dfa.states if not state == self.dfa.hole])
        return minimizable

    def _convert_to_dfa(self, eq_class, eq_class_0):
        self._remove_hole_from_eq_class(eq_class)
        new_states = []

        # Load states
        for index, partition in enumerate(eq_class):
            is_final_partition = False
            final_states = eq_class_0[0]
            if partition[0] in final_states:
                is_final_partition = True
            partition_state = State(str(index), is_final_partition)
            new_states.append(partition_state)

        # Add transitions
        for index, partition in enumerate(eq_class):
            state_of_partition = new_states[index]
            for symbol in self.dfa.alphabet.symbols:
                next_state = self._get_next_state(symbol, partition)
                if not next_state == self.dfa.hole:
                    # If its a hole, nothing needs to be done; hole transitions
                    # are added automatically on DFA construction.
                    number_of_part = self._get_num_of_partition_of_state(
                        next_state, eq_class)
                    actual_next_state = new_states[number_of_part]
                    state_of_partition.add_transition(
                        symbol, actual_next_state)

        number_of_part_of_init_state = self._get_num_of_partition_of_state(
            self.dfa.initial_state, eq_class)
        actual_initial_state = new_states[number_of_part_of_init_state]

        return DFA(self.dfa.alphabet, actual_initial_state, set(new_states), comparator=self.dfa.comparator)

    def _remove_hole_from_eq_class(self, eq_class):
        for partition in eq_class:
            for state in partition:
                if state == self.dfa.hole:
                    if len(partition) == 1:
                        eq_class.remove(partition)
                    else:
                        partition.remove(state)

    def _get_num_of_partition_of_state(self, state, eq_class):
        for index, partition in enumerate(eq_class):
            if state in partition:
                return index

    def _get_next_state(self, symbol, partition):
        for state in partition:
            next_state = next(iter(state.next_states_for(symbol)))
            if not next_state == next_state.hole:
                return next_state
        return next_state

    def _next_eq_class_from(self, equivalence_class):
        new_eq_class = []
        for partition in equivalence_class:
            if len(partition) == 1:
                # Non partitionable
                new_eq_class.append(partition)
            else:
                number_of_states = len(partition)

                not_distinguishable_table = self._create_false_filled_square_table(
                    number_of_states)
                states = [state for state in partition]

                # Skip first element
                for i in range(1, number_of_states):
                    # Only below diagonal
                    for j in range(0, i):
                        not_distinguishable_table[i][j] = not self._are_distinguishable(
                            states[i], states[j], equivalence_class)

                self._fill_diag_and_make_symetric(not_distinguishable_table)
                eq_class_from_table = self._create_eq_class_from_table(
                    not_distinguishable_table, states)
                new_eq_class.extend(eq_class_from_table)
        return new_eq_class

    def _create_eq_class_from_table(self, non_distinguishable_table, states):
        moved_state_lst = [False for row in non_distinguishable_table]
        eq_class = []
        for i, row in enumerate(non_distinguishable_table):
            if not moved_state_lst[i]:
                partition = []
                for j, non_distinguishable in enumerate(row):
                    if non_distinguishable:
                        partition.append(states[j])
                        moved_state_lst[j] = True
                if len(partition) > 0:
                    eq_class.append(partition)
        return eq_class

    def _fill_diag_and_make_symetric(self, table):
        for i in range(0, len(table)):
            for j in range(0, len(table[0])):
                if i == j:
                    table[i][j] = True
                else:
                    table[i][j] = table[j][i]

    def _are_distinguishable(self, state1, state2, equivalence_class):
        distinguishable = False
        for symbol in self.dfa.alphabet.symbols:

            next_state_1 = next(iter(state1.next_states_for(symbol)))
            next_state_2 = next(iter(state2.next_states_for(symbol)))

            partition_where_1_belongs = self._get_partition_where_it_belongs(
                next_state_1, equivalence_class)
            partition_where_2_belongs = self._get_partition_where_it_belongs(
                next_state_2, equivalence_class)

            if not partition_where_1_belongs == partition_where_2_belongs:
                distinguishable = True
                break
        return distinguishable

    def _get_partition_where_it_belongs(self, state, partitions):
        if state == self.dfa.hole:
            return None
        for partition in partitions:
            if state in partition:
                return partition

    def _create_false_filled_square_table(self, dimension: int) -> list[list[bool]]:
        return [[False for x in range(dimension)]
                for x in range(dimension)]

    def _eq_class_iterations_finished(self, equivalence_classes: list, current_iteration: int) -> bool:
        are_same_equivalence_class = len(equivalence_classes[current_iteration]) == len(
            equivalence_classes[current_iteration - 1])
        return are_same_equivalence_class

    def _get_final_states(self) -> list[State]:
        return [x for x in self.dfa.states if x.is_final and not x == self.dfa.hole]

    def _get_non_final_states(self) -> list[State]:
        return [x for x in self.dfa.states if not x.is_final]



import uuid
from typing import Any

from pythautomata.abstract.pdfa_model_exporting_strategy import PDFAModelExportingStrategy
from pythautomata.abstract.probabilistic_model import ProbabilisticModel
from pythautomata.automata.wheighted_automaton_definition.weighted_automaton import WeightedAutomaton
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import Symbol, SymbolStr
from pythautomata.model_exporters.wfa_image_exporter import WFAImageExporter
from pythautomata.abstract.finite_automaton import FiniteAutomataComparator


def is_probabilistic(weighted_states: set[WeightedState], alphabet: Alphabet, max_error=0.3) -> bool:
    initial_states = list(
        filter(lambda state: state.initial_weight != 0, weighted_states))
    if len(initial_states) != 1:
        print('Assertion Error: len(initial_states) != 1')
        return False
    for weighted_state in weighted_states:
        total_weight = weighted_state.final_weight
        for symbol in alphabet.symbols:
            transitions_for_symbol = weighted_state.transitions_list[symbol]
            if len(transitions_for_symbol) > 1:
                print('Assertion Error:len(transitions_for_symbol) > 1')
                print('weighted_state:', weighted_state.name)
                return False
            total_weight += transitions_for_symbol[0][1]
        if abs(total_weight-1.0) > max_error:
            print('Assertion Error: abs(total_weight-1.0) > max_error')
            print('weighted_state:', weighted_state.name)
            print('total_weight:', total_weight)
            return False
    return True


class ProbabilisticDeterministicFiniteAutomaton(WeightedAutomaton, ProbabilisticModel):

    def __init__(self, alphabet: Alphabet, weighted_states: set, terminal_symbol: Symbol,
                 comparator: FiniteAutomataComparator,
                 name=None,
                 export_strategies: list[PDFAModelExportingStrategy] = None):
        assert is_probabilistic(
            weighted_states, alphabet), "Trying to instantiate a non probabilisitic DFA"
        if export_strategies is None:
            export_strategies = [WFAImageExporter()]
        if name is None:
            name = 'PDFA - ' + str(uuid.uuid4().hex)
        super().__init__(alphabet, weighted_states,
                         terminal_symbol, comparator, name, export_strategies)

    def log_sequence_probability(self, sequence: Sequence) -> float:
        return self.log_sequence_weight(sequence)

    def last_token_probability(self, sequence: Sequence) -> float:
        return self.last_token_weight(sequence)[0]

    def sequence_probability(self, sequence: Sequence) -> float:
        return self.sequence_weight(sequence)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ProbabilisticDeterministicFiniteAutomaton) and self._comparator.are_equivalent(self,
                                                                                                                other)

    def get_first_state(self):
        for state in self.weighted_states:
            if state.initial_weight == 1:
                return state

    def __getstate__(self):
        lines = f'Name\n{self.name}\n'
        lines += f'Terminal Symbol\n{self.terminal_symbol}\n'
        lines += 'Alphabet\n' + \
            ' '.join(
                map(lambda symbol: f'{str(symbol)}', self.alphabet.symbols)) + '\n'
        lines += ''.join(map(self.__encode_state, self.weighted_states))
        lines += ''.join(map(self.__encode_transitions_for_state,
                         self.weighted_states))
        return lines

    def __encode_state(self, state):
        return f'State\n{state.name} {state.initial_weight} {state.final_weight}\n'

    def __encode_transitions_for_state(self, state):
        if not state.transitions_list:
            return ''
        transitions_lines = ''.join(map(lambda x: self.__encode_transitions_for_state_symbol(state.name, x),
                                        state.transitions_list.items()))
        return transitions_lines

    def __encode_transitions_for_state_symbol(self, origin, transitions):
        symbol, transition_list = transitions
        transitions_lines = ''.join(map(lambda x: self.__encode_transition(origin, symbol, x),
                                        transition_list))
        return transitions_lines

    def __encode_transition(self, origin, symbol, transition):
        next_state, weight = transition
        return f'Transition\n{origin} {next_state.name} {symbol} {weight}\n'

    def __setstate__(self, state):
        elements = {'states': dict()}
        lines = str.split(str(state), '\n')[:-1]
        for i in range(0, len(lines), 2):
            self.decode(lines[i], lines[i + 1], elements)
        self._name = elements['name']
        self._alphabet = elements['alphabet']
        self._terminal_symbol = elements['terminal_symbol']
        self.weighted_states = set(elements['states'].values())

    def decode(self, key, value, acc):
        {'Name': lambda x: self.decode_name(value, x),
         'Terminal Symbol': lambda x: self.decode_terminal_symbol(value, x),
         'Alphabet': lambda x: self.decode_alphabet(value, x),
         'State': lambda x: self.decode_state(value, x),
         'Transition': lambda x: self.decode_transition(value, x)}[key](acc)

    def decode_name(self, value, acc):
        acc['name'] = value

    def decode_terminal_symbol(self, value, acc):
        acc['terminal_symbol'] = SymbolStr(value)

    def decode_alphabet(self, value, acc):
        symbols = value.split(' ')
        acc['alphabet'] = Alphabet(frozenset(map(SymbolStr, symbols)))

    def decode_state(self, value, acc):
        state_elements = value.split(' ')
        acc['states'][state_elements[0]] = (WeightedState(state_elements[0],
                                                          float(
                                                              state_elements[1]),
                                                          float(state_elements[2])))

    def decode_transition(self, value, acc):
        transition_elements = value.split(' ')
        acc['states'][transition_elements[0]].add_transition(SymbolStr(transition_elements[2]),
                                                             acc['states'][transition_elements[1]],
                                                             float(transition_elements[3]))



from __future__ import annotations
from pythautomata.automata.wheighted_automaton_definition.weighted_transition import WeightedTransition
from pythautomata.base_types.symbol import Symbol
from pythautomata.exceptions.none_state_exception import NoneStateException
from pythautomata.exceptions.non_deterministic_states_exception import NonDeterministicStatesException


class WeightedState:

    def __init__(self, name, initial_weight, final_weight):
        self.name = name
        self.transitions_set: dict[Symbol, set[WeightedTransition]] = dict()
        self.transitions_list: dict[Symbol, list[tuple[WeightedState, float]]] = dict()
        self.initial_weight: float = initial_weight
        self.final_weight: float = final_weight

    def add_transition(self, symbol: Symbol, next_state: WeightedState, weight: float) -> None:
        if next_state is None:
            raise NoneStateException()
        if symbol not in self.transitions_set:
            self.transitions_set[symbol] = set()
            self.transitions_list[symbol] = list()
        transition = WeightedTransition(next_state, weight)
        if transition not in self.transitions_set[symbol]:
            self.transitions_set[symbol].add(WeightedTransition(next_state, weight))
            self.transitions_list[symbol].append((next_state, weight))

    def transitions_set_for(self, symbol: Symbol) -> set[WeightedTransition]:
        if symbol not in self.transitions_set:
            raise Exception(f'No transition for symbol: {symbol}')
        return self.transitions_set[symbol]

    def transitions_list_for(self, symbol) -> list[tuple[WeightedState, float]]:
        if symbol not in self.transitions_list.keys():
            raise Exception(f'No transition for symbol: {symbol}')
        return self.transitions_list[symbol]

    def next_states_for(self, symbol: Symbol) -> set['WeightedState']:
        if symbol not in self.transitions_list.keys():
            raise NonDeterministicStatesException()
        return set(t[0] for t in self.transitions_list[symbol])

    def get_all_symbol_weights(self, terminal_symbol) -> tuple[list[Symbol], list[float], list[WeightedState]]:
        symbols = list()
        weights = list()
        next_states = list()
        for symbol, weighted_transition_set in self.transitions_set.items():
            for weighted_transition in weighted_transition_set:
                symbols.append(symbol)
                weights.append(weighted_transition.weight)
                next_states.append(weighted_transition.next_state)
        weights.append(self.final_weight)
        next_states.append(None)
        symbols.append(terminal_symbol)
        return symbols, weights, next_states

    def __eq__(self, other):
        if not isinstance(other, WeightedState):
            return False
        else:
            return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return str(self.name)



from unittest import TestCase
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.sequence import Sequence

abcAlphabet = Alphabet(frozenset(
    [SymbolStr('a'), SymbolStr('b'), SymbolStr('c')]))


class TestSequence(TestCase):

    def test_word(self):
        sequence = Sequence([SymbolStr('a'), SymbolStr('b'), SymbolStr('c')])
        self.assertEqual(len(sequence.value), 3)

    def test_word_iteration(self):
        sequence = Sequence([SymbolStr('a'), SymbolStr('b'), SymbolStr('c')])
        sequence += SymbolStr('b')
        self.assertEqual(len(sequence.value), 4)

    def test_word_iteration_2(self):
        sequence = Sequence([SymbolStr('a'), SymbolStr('b'), SymbolStr('c')])
        sequence += Sequence([SymbolStr('b'), SymbolStr('c')])
        self.assertEqual(len(sequence.value), 5)



from typing import Tuple

from pythautomata.base_types.guard import Guard
from pythautomata.base_types.symbol import Symbol
from pythautomata.exceptions.none_state_exception import NoneStateException


class SymbolicState:
    """Representation of SFA states.

        Attributes
        ----------
        name: str
            State name.
        is_final: bool
            Determines if the state is final.
        transitions: list[Tuple[Guard,SymbolicState]]
            A list containing guards associated with the next state
        hole:
            Hole state, state containing all transitions directed to itself. 
            It is used as default when a symbol is not present as transition key.
    """

    def __init__(self, name: str, is_final: bool = False):
        self.name = name
        self.is_final = is_final
        self.transitions: list[Tuple[Guard, SymbolicState]] = []

    @property
    def hole(self) -> 'SymbolicState':
        return self._hole

    def add_transition(self, guard: Guard, next_state: 'SymbolicState') -> None:
        """Adds a transition consisting of a guard and the next state

        Args:
            guard (Guard): A guard contains a predicate that receives a symbol
            next_state (SymbolicState): The state to go to when the guard is true

        Raises:
            NoneStateException: when next_state is None this exception is raised
        """
        if next_state is None:
            raise NoneStateException()
        self.transitions.append((guard, next_state))

    def next_state_for(self, symbol: Symbol) -> 'SymbolicState':
        for guard, state in self.transitions:
            if guard.matches(symbol):
                return state
        return self._hole

    def next_states_for(self, symbol: Symbol) -> list['SymbolicState']:
        return [self.next_state_for(symbol)]

    def add_hole_transition(self, hole: 'SymbolicState') -> None:
        self._hole = hole

    def __eq__(self, other):
        return isinstance(other, SymbolicState) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return str(self.name) + (" (Final)" if self.is_final else " (Non-final)")



from unittest import TestCase

from pythautomata.other_models.probabilistic_filter_model import ProbabilisticFilterModel
from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.symbol import SymbolStr


class TestProbabilisticFilterModel(TestCase):

    def test_1(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        model = ProbabilisticFilterModel(pdfa, 2)
        seq1 = Sequence([])
        seq2 = Sequence([SymbolStr('0'), SymbolStr('1')])
        seq3 = Sequence([SymbolStr('0'), SymbolStr('1'), SymbolStr('$')])
        
        pdfa_res1 = pdfa.last_token_probability(seq1)        
        pdfa_res2 = pdfa.last_token_probability(seq2)        
        pdfa_res3 = pdfa.last_token_probability(seq3)

        model_res1 = model.last_token_probability(seq1)        
        model_res2 = model.last_token_probability(seq2)        
        model_res3 = model.last_token_probability(seq3)

        self.assertEqual(pdfa_res1,model_res1)
        self.assertEqual(pdfa_res2,model_res2)
        self.assertNotEqual(pdfa_res3, model_res3)


from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton as DFA
from pythautomata.automata.non_deterministic_finite_automaton import \
    NondeterministicFiniteAutomaton as NFA
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.model_comparators.dfa_comparison_strategy import \
    DFAComparisonStrategy as DFAComparator
from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import \
    HopcroftKarpComparisonStrategy as HopcroftKarpComparison

abAlphabet = Alphabet(frozenset((SymbolStr('a'), SymbolStr('b'))))
abcAlphabet = Alphabet(
    frozenset((SymbolStr('a'), SymbolStr('b'), SymbolStr('c'))))

# TODO: Check if some NFAS are not in the paper and test


class BolligHabermehlKernLeuckerAutomata:
    """
    Class containing automata from paper:
        Bollig B. and Habermehl P. and Kern C. and Leucker M. 2009. 
        Angluin-style learning of NFA. 
        IJCAI International Joint Conference on Artificial Intelligence. 1004-1009. 

    Methods
    -------
    Methods
    -------
    get_all_automata: list(FiniteAutomaton)
        returns a list containing all automata defined in this class

    get_first_example_DFA: DeterministicFiniteAutomaton
        returns the automaton 1 from the paper 

    get_first_example_NFA: NondeterministicFiniteAutomaton
        returns the automaton 2 from the paper

    """
    @staticmethod
    def get_all_automata():
        """
        Method returning a list of all automata of the class

        Returns
        -------
        list(FiniteAutomaton)
            all automata defined in the class
        """
        return [
            BolligHabermehlKernLeuckerAutomata.get_first_example_DFA(),
            BolligHabermehlKernLeuckerAutomata.get_first_example_NFA()
        ]

    @staticmethod
    def get_first_example_DFA() -> DFA:
        """
        Method returning a DFA that recognizes the regular expresion Σ*aΣ^2 (First example from paper, Fig. 3.)

        Returns:
            DFA: Method returning a DFA that recognizes the regular expresion Σ*aΣ^2
        """
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("State 0")
        state1 = State("State 1")
        state2 = State("State 2")
        state3 = State("State 3", True)
        state4 = State("State 4", True)
        state5 = State("State 5", True)
        state6 = State("State 6")
        state7 = State("State 7", True)

        state0.add_transition(a, state1)
        state0.add_transition(b, state0)
        state1.add_transition(a, state2)
        state1.add_transition(b, state6)
        state2.add_transition(a, state3)
        state2.add_transition(b, state4)
        state3.add_transition(a, state3)
        state3.add_transition(b, state4)
        state4.add_transition(a, state5)
        state4.add_transition(b, state7)
        state5.add_transition(a, state2)
        state5.add_transition(b, state6)
        state6.add_transition(a, state5)
        state6.add_transition(b, state7)
        state7.add_transition(a, state1)
        state7.add_transition(b, state0)
        comparator = DFAComparator()
        return DFA(abAlphabet, state0,
                   set([state0, state1, state2, state3,
                        state4, state5, state6, state7]),
                   comparator,
                   "Angluin-style learning of NFA - first example DFA")

    @staticmethod
    def get_first_example_NFA() -> NFA:
        """
        Method returning a NFA that recognizes the regular expresion Σ*aΣ^2 (First example from paper, Fig. 4.)

        Returns:
            NFA: Method returning a NFA that recognizes the regular expresion Σ*aΣ^2
        """
        a = abAlphabet['a']
        b = abAlphabet['b']

        state0 = State("-+--")
        state1 = State("-++-")
        state2 = State("-+-+")
        state3 = State("++--", True)

        state0.add_transition(a, state1)
        state0.add_transition(a, state0)
        state0.add_transition(b, state0)
        state1.add_transition(a, state1)
        state1.add_transition(a, state0)
        state1.add_transition(b, state0)
        state1.add_transition(a, state2)
        state1.add_transition(b, state2)
        state2.add_transition(a, state1)
        state2.add_transition(a, state0)
        state2.add_transition(b, state0)
        state2.add_transition(a, state3)
        state2.add_transition(b, state3)
        state3.add_transition(a, state1)
        state3.add_transition(a, state0)
        state3.add_transition(b, state0)

        comparator = HopcroftKarpComparison()
        return NFA(abAlphabet,  frozenset([state0]),
                   set([state0, state1, state2, state3]),
                   comparator,
                   "Angluin-style learning of NFA - first example NFA")



import numpy as np
from random import seed
from random import randint

from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
from typing import List


class SequenceGenerator:

    def __init__(self, alphabet: Alphabet, max_seq_length: int, random_seed: int = 21, min_seq_length: int = 0):
        self._alphabet = alphabet
        self._min_seq_length = min_seq_length
        self._max_seq_length = max_seq_length
        self._seed = random_seed
        seed(self._seed)

    def reset_seed(self):
        seed(self._seed)

    def generate_word(self, length):
        if length > self._max_seq_length:
            raise AssertionError("Param length cannot exceed max_seq_length")

        value = []
        list_symbols = list(self._alphabet.symbols)
        list_symbols.sort()
        for _ in range(length):
            position = randint(0, len(list_symbols) - 1)
            symbol = list_symbols[position]
            value.append(symbol)
        return Sequence(value)

    def generate_words(self, number_of_words: int):
        result = np.empty(number_of_words, dtype=Sequence)
        for index in range(number_of_words):
            length = randint(self._min_seq_length, self._max_seq_length)
            result[index] = self.generate_word(length)
        return result

    def generate_words_with_padding(self, number_of_words: int, padding_symbol: Symbol):
        result = np.empty(number_of_words, dtype=Sequence)
        for index in range(number_of_words):
            length = randint(self._min_seq_length, self._max_seq_length)
            word = self.generate_word(length)
            result[index] = self.pad(word, padding_symbol)
        return result

    def pad(self, word: Sequence, padding_symbol: Symbol, max_len=None, padding_type='post'):
        if max_len is None:
            max_len = self._max_seq_length

        value = list(word.value)
        if len(value) > max_len:
            if padding_type == 'post':
                value = value[0:max_len]
            elif padding_type == 'pre':
                value = value[len(value) - max_len:len(value)]

        while len(value) < max_len:
            if padding_type == 'post':
                value.append(padding_symbol)
            elif padding_type == 'pre':
                value = [padding_symbol] + value
            else:
                raise Exception('Unknown padding type')
        return Sequence(value)

    def pad_sequences(self, words: List[Sequence], padding_symbol: Symbol, max_len=None, padding_type='post'):
        padded_sequences = list(map(lambda x: self.pad(x, padding_symbol, max_len, padding_type), words))
        return padded_sequences

    def generate_all_words_up_to_max_length(self):
        ret = [Sequence([])]
        list_symbols = list(self._alphabet.symbols)
        list_symbols.sort()
        for counter in range(self._max_seq_length):
            ret_aux = ret.copy()
            for i in range(len(ret)):
                for symbol in list_symbols:
                    if len(ret[i].value) >= counter:
                        value = list(ret[i].value)
                        value.append(symbol)
                        extension = Sequence(value)
                        ret_aux.append(extension)
            ret = ret_aux
        return ret



from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA
from pythautomata.model_comparators.dfa_comparison_strategy import DFAComparisonStrategy as AutomataComparator
from pythautomata.base_types.state import State

#TODO: ADD DOCSTRINGS

def join_DFAs(dfa1: DFA, dfa2: DFA) -> DFA:
    """
    Function implementing DFA join. 
    The join between two DFAs is a new DFA that recognizes only words belonging to languages of both DFAs at the same time.

    Args:
        dfa1 (DFA): A DFA.
        dfa2 (DFA): Another DFA.

    Returns:
        DFA: A DFA containing words that are accepted by dfa1 and dfa2 at the same time.
    """
    _check_alphabets(dfa1, dfa2)  

    initial_pair = _get_initial_pair(dfa1, dfa2)
    join_DFA_initial_state =  _pair_to_state_join(initial_pair)    
    join_DFA_alphabet = dfa1.alphabet

    pairs_to_visit = [initial_pair]
    visited_pairs = set()
    result_states = dict()
    result_states[join_DFA_initial_state.name]= join_DFA_initial_state
    while len(pairs_to_visit) > 0:
        pair = pairs_to_visit[0]
        new_state = _pair_to_state_join(pair)
        
        if(new_state.name not in result_states.keys()):
            result_states[new_state.name]= new_state
        else:
            new_state = result_states[new_state.name]

        for symbol in join_DFA_alphabet.symbols:
            next_pair = (pair[0].next_state_for(
                symbol), pair[1].next_state_for(symbol))

            next_state = _pair_to_state_join(next_pair)

            if next_pair not in pairs_to_visit and next_pair not in visited_pairs:                    
                pairs_to_visit.append(next_pair)
                result_states[next_state.name] = next_state
            else:
                next_state = result_states[next_state.name]  

            new_state.add_transition(symbol,next_state)
        
        pairs_to_visit.remove(pair)
        visited_pairs.add(pair)

    #Update the initial state, so it has the transitions
    join_DFA_initial_state = result_states[join_DFA_initial_state.name]
    join_DFA_states = set(result_states.values())
    comparator = AutomataComparator()
    resulting_DFA = DFA(join_DFA_alphabet,join_DFA_initial_state,join_DFA_states, comparator = comparator)
    return resulting_DFA

def union_DFAs(dfa1: DFA, dfa2: DFA) -> DFA:
    """
    Function implementing DFA union. 
    The union between two DFAs is a new DFA that recognizes words belonging to languages of any of the DFAs.

    Args:
        dfa1 (DFA): A DFA.
        dfa2 (DFA): Another DFA.

    Returns:
        DFA: A DFA containing words that are accepted by dfa1 or dfa2.
    """
    _check_alphabets(dfa1, dfa2)    

    initial_pair = _get_initial_pair(dfa1, dfa2)
    union_DFA_initial_state =  _pair_to_state_union(initial_pair)    
    union_DFA_alphabet = dfa1.alphabet

    pairs_to_visit = [initial_pair]
    visited_pairs = set()
    result_states = dict()
    result_states[union_DFA_initial_state.name]= union_DFA_initial_state
    while len(pairs_to_visit) > 0:
        pair = pairs_to_visit[0]
        new_state = _pair_to_state_union(pair)
        
        if(new_state.name not in result_states.keys()):
            result_states[new_state.name]= new_state
        else:
            new_state = result_states[new_state.name]

        for symbol in union_DFA_alphabet.symbols:
            next_pair = (pair[0].next_state_for(
                symbol), pair[1].next_state_for(symbol))

            next_state = _pair_to_state_union(next_pair)

            if next_pair not in pairs_to_visit and next_pair not in visited_pairs:                    
                pairs_to_visit.append(next_pair)
                result_states[next_state.name] = next_state
            else:
                next_state = result_states[next_state.name]  

            new_state.add_transition(symbol,next_state)
        
        pairs_to_visit.remove(pair)
        visited_pairs.add(pair)

    #Update the initial state, so it has the transitions
    union_DFA_initial_states = result_states[union_DFA_initial_state.name]
    union_DFA_states = set(result_states.values())
    comparator = AutomataComparator()
    resulting_DFA = DFA(union_DFA_alphabet,union_DFA_initial_states,union_DFA_states, comparator = comparator)
    return resulting_DFA

def join_DFA_set(dfaSet: set[DFA]) -> DFA:
    """
    Function implementing DFA join of a whole set of DFAs. 
    The join between DFAs is a new DFA that recognizes words belonging to languages of all of the DFAs at the same time.

    Args:
        dfaSet (set[DFA]): A set of DFAs.

    Returns:
        DFA: A DFA containing words that are accepted by all of the DFAs in the set.
    """
    listDFAs = list(dfaSet)
    result = listDFAs[0]
    for dfa in listDFAs:
        result = join_DFAs(result, dfa)
    return result    

def union_DFA_set(dfaSet: set[DFA]) -> DFA:
    """
    Function implementing DFA union of a whole set of DFAs. 
    The union between DFAs is a new DFA that recognizes words belonging to languages of any of the DFAs.

    Args:
        dfaSet (set[DFA]): A set of DFAs.

    Returns:
        DFA: A DFA containing words that are accepted by any of the DFAs in the set.
    """

    listDFAs = list(dfaSet)
    result = listDFAs[0]
    for dfa in listDFAs:
        result = union_DFAs(result, dfa)
    return result   

def _get_initial_pair(dfa1: DFA, dfa2:DFA) -> tuple[State, State]:
    initial_state1 = dfa1.initial_state
    initial_state2 = dfa2.initial_state
    return (initial_state1, initial_state2)    

def _check_alphabets(dfa1: DFA, dfa2:DFA) -> None:
    if dfa1.alphabet != dfa2.alphabet:
            raise ValueError("Alphabets are not equivalent.")   

def _pair_to_state_join(pair: tuple[State, State]) -> State:
    state0 = pair[0]    
    state1 = pair[1]
    new_state_is_final = state0.is_final and state1.is_final
    new_state_name = state0.name+state1.name
    new_state = State(new_state_name,new_state_is_final)
    return new_state
   
def _pair_to_state_union(pair: tuple[State, State]) -> State:
    state0 = pair[0]    
    state1 = pair[1]
    new_state_is_final = state0.is_final or state1.is_final
    new_state_name = state0.name+state1.name
    new_state = State(new_state_name,new_state_is_final)
    return new_state    




from graphviz import Digraph
from os import makedirs
from os.path import isdir
from pathlib import Path


class StandardDotMMExportingStrategy:

    def export(self, model, path=None, encoding=None):
        graph = self.create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding=encoding) as f:
            f.write(dot_code)

    def create_graph(self, model):
        graph = Digraph('moore-machine-automata')

        graph.node('__start0', shape='none', label='')
        initialState = model.initial_state
        graph.edge('__start0', initialState.name)

        for state in model.states:
            label = "{ %s | %s }" % (state.name, state.value.value)
            graph.node(state.name, shape='record', style='rounded', label=label)

        for state in model.states:
            for symbol, destinationStates in state.transitions.items():
                for destinationState in destinationStates:
                    graph.edge(state.name, destinationState.name, str(symbol))

        return graph

    def get_path_for(self, path: str, model):
        if path is None:
            name = model._name
            path = "output_models/" + \
                ("" if name is None else f"{name}")
        if not isdir(path):
            makedirs(path)
        return Path(path, "Standard %s" % (model._name))



from pythautomata.guards.equality_guard import EqualityGuard
from pythautomata.guards.union_guard import UnionGuard
from pythautomata.base_types.guard import Guard
from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.symbolic_state import SymbolicState
from pythautomata.boolean_algebra_learner.boolean_algebra_learner import BooleanAlgebraLearner


class EqualityLearner(BooleanAlgebraLearner):
    def __init__(self):
        return

    def learn(self, multidict: dict[SymbolicState, list[Symbol]]) -> list[tuple[Guard, SymbolicState]]:
        ret:list[tuple[Guard, SymbolicState]] = []

        symbols:list[Symbol]
        for state, symbols in multidict.items():
            guard = UnionGuard(*[EqualityGuard(symbol) for symbol in symbols])
            ret.append((guard, state))
            
        return ret



from unittest import TestCase

from pythautomata.automata_definitions.other_automata import OtherAutomata
from pythautomata.utilities.dfa_minimizer import DFAMinimizer


class TestMinimizer(TestCase):

    def test_minimization_1(self):
        automaton = OtherAutomata.get_automaton_1()
        minimized_automaton = DFAMinimizer(automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_automaton_1_minimized()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_2(self):
        non_minimizable_automaton = OtherAutomata.get_non_minimizable_automaton_1()
        minimized_automaton = DFAMinimizer(
            non_minimizable_automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_non_minimizable_automaton_1()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_3(self):
        non_minimizable_automaton = OtherAutomata.get_non_minimizable_automaton_2()
        minimized_automaton = DFAMinimizer(
            non_minimizable_automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_non_minimizable_automaton_2()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_4(self):
        automaton = OtherAutomata.get_automaton_2()
        minimized_automaton = DFAMinimizer(automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_automaton_2_minimized()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_5(self):
        automaton = OtherAutomata.get_automaton_3()
        minimized_automaton = DFAMinimizer(automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_automaton_3_minimized()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_6(self):
        non_minimizable_automaton = OtherAutomata.get_non_minimizable_automaton_3()
        minimized_automaton = DFAMinimizer(
            non_minimizable_automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_non_minimizable_automaton_3()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)

    def test_minimization_7(self):
        automaton = OtherAutomata.get_automaton_4()
        minimized_automaton = DFAMinimizer(automaton).minimize()
        expected_minimized_automaton = OtherAutomata.get_automaton_4_minimized()
        self.assertEqual(minimized_automaton, expected_minimized_automaton)



from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton as DFA
import pythautomata.utilities.simple_dfa_generator as simple_dfa_generator
import math
from scipy.special import lambertw
from math import exp
from pythautomata.model_exporters.dot_exporting_strategy import DotExportingStrategy


def generate_dfa(alphabet: Alphabet, nominal_size: int, seed: int = None, exporting_strategies: list = [DotExportingStrategy()]) -> DFA:
    """
    Function returning a randomly generated DFA in the manner descbribed by Cyril Nicaud in: 
        Nicaud, Cyril. (2014). Random Deterministic Automata. 5-23. 10.1007/978-3-662-44522-8_2. 

    Args:
        alphabet (Alphabet): DFA alphabet.
        nominal_size (int): Target nominal size of the generated DFA.
        seed (int, optional): Random seed. Defaults to None.
        exporting_strategies (list, optional): List of exporting strategies. Defaults to [DotExportingStrategy()].

    Returns:
        DFA: Random DFA with distribution of sizes centered near nominal_size and depth centered near 2*log(nominal_size,2)-2
    """
    k = len(alphabet)
    ro_k = k + lambertw(-k * exp(-k))
    ro_k = ro_k.real
    v_k = ro_k/k
    number_of_states = math.ceil(nominal_size / v_k)
    dfa = simple_dfa_generator.generate_dfa(
        alphabet, number_of_states, seed, exporting_strategies)
    return dfa



from pythautomata.abstract.probabilistic_model import ProbabilisticModel
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.sequence import Sequence


class ProbabilisticFilterModel(ProbabilisticModel):

    def __init__(self, model: ProbabilisticModel, max_length: int):
        self._model = model        
        alphabet_symbols = set(self._model.alphabet.symbols)
        #filter_symbol = SymbolStr('#')
        #assert(filter_symbol not in alphabet_symbols)
        #alphabet_symbols.add(filter_symbol)
        self._alphabet = Alphabet(alphabet_symbols)
        self._max_length = max_length
        #self._filter_symbol = filter_symbol
    
    @property
    def name(self) -> str:
        return self._model.name+"_filtered"

    @property
    def terminal_symbol(self) -> SymbolStr:
        return self._model.terminal_symbol
        
    @property
    def alphabet(self) -> Alphabet:
        return self._alphabet

    def sequence_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    def log_sequence_probability(self, sequence: Sequence) -> float:
        raise NotImplementedError

    def last_token_probability(self, sequence: Sequence) -> float:
        if len(sequence) > self._max_length:
            if sequence[-1] == self._model.terminal_symbol:
                return 1.0
            else:
                return 0.0
        else:
            return self._model.last_token_probability(sequence)


   



from unittest import TestCase

from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.utilities import pdfa_generator


class TestPDFAGenerator(TestCase):

    def test_1(self):
        dfa = TomitasGrammars.get_automaton_1()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_2(self):
        dfa = TomitasGrammars.get_automaton_2()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_3(self):
        dfa = TomitasGrammars.get_automaton_3()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_4(self):
        dfa = TomitasGrammars.get_automaton_4()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_5(self):
        dfa = TomitasGrammars.get_automaton_5()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_6(self):
        dfa = TomitasGrammars.get_automaton_6()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_7(self):
        dfa = TomitasGrammars.get_automaton_7()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa)
        self.assertEqual(pdfa, pdfa)

    def test_8(self):
        dfa = TomitasGrammars.get_automaton_7()
        pdfa = pdfa_generator.pdfa_from_dfa(dfa, distributions=3, max_shift=0)
        self.assertEqual(pdfa, pdfa)

    def test_9(self):
        dfa = TomitasGrammars.get_automaton_7()
        pdfa = pdfa_generator.pdfa_from_dfa(
            dfa, distributions=1, max_shift=0.1)
        self.assertEqual(pdfa, pdfa)



from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy
from pythautomata.automata.non_deterministic_finite_automaton import NondeterministicFiniteAutomaton
from pythautomata.base_types.symbolic_state import SymbolicState
from graphviz import Digraph


class StandardDotExportingStrategy(ModelExportingStrategy):

    def export(self, model, path=None, encoding=None):
        graph = self.create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding=encoding) as f:
            f.write(dot_code)

    def create_graph(self, model):
        if type(model) == NondeterministicFiniteAutomaton:
            raise NotImplemented

        graph = Digraph('finite_state_machine')

        graph.node('__start0', shape='none', label='')
        for state in model.states:
            if state.is_final:
                graph.node(state.name, label=state.name, shape='doublecircle')
            else:
                graph.node(state.name, label=state.name, shape='circle')

        initialState = model.initial_state
        graph.edge('__start0', initialState.name, label='')
        for state in model.states:
            for symbol, destinationStates in state.transitions.items():
                for destinationState in destinationStates:
                    graph.edge(state.name, destinationState.name, str(symbol))

        return graph



from .python_automata_exception import PythonAutomataException

class NonDeterministicStatesException(PythonAutomataException):
    def __init__(self):
        self.message = "All states must be deterministic in order to create a Deterministic Finite Automaton"


from unittest import TestCase
from pythautomata.utilities import simple_dfa_generator
from pythautomata.utilities import abbadingo_one_dfa_generator
from pythautomata.utilities import nicaud_dfa_generator
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr


class TestDFAGenerators(TestCase):

    def test_simple_dfa_generator_1(self):
        binaryAlphabet = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1')]))
        generated_automata = simple_dfa_generator.generate_dfa(
            binaryAlphabet, 80)
        self._assert_correctness(generated_automata)

    def test_simple_dfa_generator_2(self):
        binaryAlphabet = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1')]))
        generated_automata = simple_dfa_generator.generate_dfa(
            binaryAlphabet, 49)
        self._assert_correctness(generated_automata)

    def test_simple_dfa_generator_3(self):
        abcdAlphabet = Alphabet(frozenset(
            [SymbolStr('a'), SymbolStr('b'), SymbolStr('c'), SymbolStr('d')]))
        generated_automata = simple_dfa_generator.generate_dfa(
            abcdAlphabet, 49)
        self._assert_correctness(generated_automata)

    def test_simple_dfa_generator_4(self):
        alphabet012 = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1'), SymbolStr('2')]))
        generated_automata = simple_dfa_generator.generate_dfa(alphabet012, 1)
        self._assert_correctness(generated_automata)

    def test_abbadingo_one_1_excep(self):
        alphabet012 = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1'), SymbolStr('2')]))
        exc = AssertionError
        self.assertRaises(
            exc, abbadingo_one_dfa_generator.generate_dfa, alphabet012, 10)

    def test_abbadingo_one_1_ok(self):
        alphabet12 = Alphabet(
            frozenset([SymbolStr('1'), SymbolStr('2')]))
        generated_automata = abbadingo_one_dfa_generator.generate_dfa(
            alphabet12, 10)
        self._assert_correctness(generated_automata)

    def test_nicaud_1_ok(self):
        symbols = []
        for i in range(10):
            symbols.append(SymbolStr(str(i)))
        mediumAlphabet = Alphabet(frozenset(symbols))
        generated_automata = nicaud_dfa_generator.generate_dfa(
            mediumAlphabet, 1000)
        self._assert_correctness(generated_automata)

    def _assert_correctness(self, automaton):
        self.assertTrue(self._all_states_are_rechable(automaton))

    def _all_states_are_rechable(self, automaton):
        unrechable = automaton.states.copy()
        for state in unrechable.copy():
            for destinations in state.transitions.values():
                unrechable = unrechable - destinations
        return len(unrechable) <= 1  # Hole

    def _has_final_state(self, automaton):
        return any(state.is_final for state in automaton.states)



from pythautomata.model_comparators.wfa_comparison_strategy import WFAComparator
from pythautomata.utilities import pdfa_utils


class WFAQuantizationComparator(WFAComparator):
    """
    Class containing a WFA Comparator Strategy based on next symbol distributions and a quantization parameter.

    Methods
    -------   
    are_equivalent: bool
        returns true iif wfa1 is equivalent to wfa2 according to which partitions the distributions over next symbols belong.

    get_counterexample_between: Sequence
        returns a Sequence where the next token weights differ (or belong to different partitions)
    """

    def __init__(self, partitions: float) -> None:
        super().__init__()
        self.partitions = partitions
        self._partition_cache = dict()

    def equivalent_output(self, observation1, observation2) -> bool:
        assert(len(observation1) == len(observation2))
        for i in range(len(observation1)):
            if self._get_partition(observation1[i])!=self._get_partition(observation2[i]):
                return False
        return True

    def equivalent_values(self, value1, value2):
        part1 = self._get_partition(value1)
        part2 = self._get_partition(value2)
        return part1 - part2 == 0

    def _get_partition(self, value):
        if value in self._partition_cache:
            return self._partition_cache[value]
        else:
            partition = pdfa_utils.get_partition(value, self.partitions)
            self._partition_cache[value] = partition
            return partition



from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.guard import Guard


class NegationGuard(Guard):
    def __init__(self, guard: Guard):
        self.guard = guard

    def matches(self, s: Symbol):
        return not self.guard.matches(s)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (u"\u00AC(") + str(self.guard) + ")"



from unittest import TestCase
from pythautomata.utilities.regex_generator import RegularExpressionGenerator
from pythautomata.utilities.sequence_generator import SequenceGenerator
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr

class TestRegexGenerator(TestCase):

    def test_generated_correctly_1(self):
        binaryAlphabet = Alphabet(
            frozenset([SymbolStr('0'), SymbolStr('1')]))            
        generator = RegularExpressionGenerator(alphabet=binaryAlphabet)

        generated_regex = generator.generate_regular_expression_with(iterations = 10)
        self._assert_correctness(generated_regex, binaryAlphabet)
    
    def test_generated_correctly_2(self):
        symbols = []
        for i in range(20):
            symbols.append(SymbolStr(str(i)))
        bigAlphabet = Alphabet(frozenset(symbols))            
        generator = RegularExpressionGenerator(alphabet=bigAlphabet)
        generated_regex = generator.generate_regular_expression_with(iterations = 100)
        self._assert_correctness(generated_regex, bigAlphabet)

    
    def _assert_correctness(self, regex, alphabet):       
        sequence_generator = SequenceGenerator(alphabet, 100)
        for i in range(1,100):
            seq = sequence_generator.generate_word(i)
            regex.accepts(seq) 
            self.assertTrue(True)

